/***
 * Fractal ADL Parser
 * Copyright (C) 2008 STMicroelectronics
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Contact: fractal@objectweb.org
 *
 * Author: Matthieu Leclercq
 */

package org.objectweb.fractal.adl.io;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.lang.reflect.Modifier;

import org.objectweb.fractal.adl.Node;

/**
 * An {@link ObjectOutputStream} that allows to serialize graph of AST node.
 * This output stream differs from traditional {@link ObjectOutputStream} from
 * the fact that information about node class are also serialized along side
 * node data, so that node classes can be regenerated by
 * {@link ObjectInputStream}.
 * 
 * @see NodeInputStream
 */
public class NodeOutputStream extends ObjectOutputStream {

  /**
   * @see ObjectOutputStream#ObjectOutputStream(OutputStream)
   */
  public NodeOutputStream(final OutputStream out) throws IOException {
    super(out);
  }

  @Override
  protected void annotateClass(final Class<?> cl) throws IOException {
    if (Node.class.isAssignableFrom(cl)
        && !Modifier.isAbstract(cl.getModifiers())) {
      /*
       * given class is a sub-class of Node. Write node type and name of
       * interfaces implemented by the node
       */
      writeBoolean(true);
      writeUTF(cl.getName());
      writeUTF(cl.getSuperclass().getName());
      Node instance;
      try {
        instance = cl.asSubclass(Node.class).newInstance();
      } catch (final InstantiationException e) {
        throw (IOException) new IOException("Can't determine AST node type")
            .initCause(e);
      } catch (final IllegalAccessException e) {
        throw (IOException) new IOException("Can't determine AST node type")
            .initCause(e);
      }
      writeUTF(instance.astGetType());
      final Class<?>[] nodeInterfaces = cl.getInterfaces();
      // assumes that there is less than 256 interfaces
      writeByte(nodeInterfaces.length);
      for (final Class<?> itf : nodeInterfaces) {
        writeUTF(itf.getName());
      }
    } else {
      writeBoolean(false);
    }
  }
}
